"""
This subpackage contains all classes that host processes (e.g. dagit)
use to manipulate and represent definitions that are resident
in user processes and containers.  e.g. ExternalPipeline.

It also contains classes that represent historical representations
that have been persisted. e.g. HistoricalPipeline
"""

from .external import (
    ExternalExecutionPlan as ExternalExecutionPlan,
    ExternalPartitionSet as ExternalPartitionSet,
    ExternalPipeline as ExternalPipeline,
    ExternalRepository as ExternalRepository,
    ExternalSchedule as ExternalSchedule,
    ExternalSensor as ExternalSensor,
)
from .external_data import (
    ExternalExecutionParamsData as ExternalExecutionParamsData,
    ExternalExecutionParamsErrorData as ExternalExecutionParamsErrorData,
    ExternalJobRef as ExternalJobRef,
    ExternalPartitionConfigData as ExternalPartitionConfigData,
    ExternalPartitionExecutionErrorData as ExternalPartitionExecutionErrorData,
    ExternalPartitionNamesData as ExternalPartitionNamesData,
    ExternalPartitionSetData as ExternalPartitionSetData,
    ExternalPartitionSetExecutionParamData as ExternalPartitionSetExecutionParamData,
    ExternalPartitionTagsData as ExternalPartitionTagsData,
    ExternalPipelineData as ExternalPipelineData,
    ExternalPipelineSubsetResult as ExternalPipelineSubsetResult,
    ExternalPresetData as ExternalPresetData,
    ExternalRepositoryData as ExternalRepositoryData,
    ExternalScheduleData as ExternalScheduleData,
    ExternalScheduleExecutionErrorData as ExternalScheduleExecutionErrorData,
    ExternalSensorExecutionErrorData as ExternalSensorExecutionErrorData,
    ExternalTargetData as ExternalTargetData,
    external_pipeline_data_from_def as external_pipeline_data_from_def,
    external_repository_data_from_def as external_repository_data_from_def,
)
from .handle import (
    JobHandle as JobHandle,
    RepositoryHandle as RepositoryHandle,
)
from .historical import HistoricalPipeline as HistoricalPipeline
from .origin import (
    IN_PROCESS_NAME as IN_PROCESS_NAME,
    ExternalInstigatorOrigin as ExternalInstigatorOrigin,
    ExternalPipelineOrigin as ExternalPipelineOrigin,
    ExternalRepositoryOrigin as ExternalRepositoryOrigin,
    GrpcServerRepositoryLocationOrigin as GrpcServerRepositoryLocationOrigin,
    InProcessRepositoryLocationOrigin as InProcessRepositoryLocationOrigin,
    ManagedGrpcPythonEnvRepositoryLocationOrigin as ManagedGrpcPythonEnvRepositoryLocationOrigin,
    RepositoryLocationOrigin as RepositoryLocationOrigin,
)
from .pipeline_index import PipelineIndex as PipelineIndex
from .repository_location import (
    GrpcServerRepositoryLocation as GrpcServerRepositoryLocation,
    InProcessRepositoryLocation as InProcessRepositoryLocation,
    RepositoryLocation as RepositoryLocation,
)
from .represented import RepresentedPipeline as RepresentedPipeline
